"""This is a relic from before the ``__aiter__`` protocol was established; it
will be phased out before aiocoap 1.0 is released."""

import abc
import enum
import asyncio

class AsyncIterable(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    @asyncio.coroutine
    def can_peek(self):
        """Return True when a result is ready to be fetched with .get_nowait(),
        and False when no more items can be fetched."""

    @abc.abstractmethod
    @asyncio.coroutine
    def get_nowait(self):
        """Fetch the next item. This must only be called once after can_peek
        has returned True."""

class QueueWithEnd(AsyncIterable):
    """A QueueWithEnd shares a Queue's behavior in that it gets fed with put
    and consumed with get_nowait. Contrary to a Queue, this is designed to be
    consumed only by one entity, which uses the coroutine can_peek to make sure
    the get_nowait will succeed.

    Another difference between a Queue and a QueueWithEnd is that the latter
    can also terminate (which is indicated by can_peek returning False and set
    by the finish coroutine) and raise exceptions (which raise from the
    get_nowait function and are set by the put_exception coroutine).
    """
    Type = enum.Enum("QueueWithEnd.Type", "notpeeked value end exception")

    def __init__(self, maxsize=0):
        # (type, value)
        self._queue = asyncio.Queue(maxsize)
        self._ended = False
        self._flag = self.Type.notpeeked

    def __repr__(self):
        return "<%s %#x flag %s%s>" % (type(self).__name__, id(self), self._flag, " (%s)" %
                self._value if self._flag in (self.Type.value,
                    self.Type.exception) else "")

    # AsyncIterable interface

    @asyncio.coroutine
    def can_peek(self):
        if self._flag is not self.Type.notpeeked:
            return True
        self._flag, self._value = yield from self._queue.get()
        return self._flag is not self.Type.end

    def get_nowait(self):
        if self._flag in (self.Type.notpeeked, self.Type.end):
            raise asyncio.QueueEmpty()
        elif self._flag is self.Type.exception:
            raise self._value
        else:
            self._flag = self.Type.notpeeked
            return self._value

    # feeder interface

    @asyncio.coroutine
    def put(self, value):
        yield from self._put(self.Type.value, value)

    @asyncio.coroutine
    def put_exception(self, value):
        yield from self._put(self.Type.exception, value)
        self._ended = True

    @asyncio.coroutine
    def finish(self):
        yield from self._put(self.Type.end, None)
        self._ended = True

    @asyncio.coroutine
    def _put(self, type, value):
        if self._ended:
            raise asyncio.InvalidStateError("%s has already ended"%type(self).__name__)
        yield from self._queue.put((type, value))

    # a simple way to create a feeder with something like an explicit yield

    @classmethod
    def cogenerator(cls, maxsize=0):
        """Coroutine decorator that passes a callable `asyncyield` into the function
        as the first argument and returns a QueueWithEnd. It is implicitly
        finished when the coroutine returns.

        >>> @QueueWithEnd.cogenerator()
        >>> def count_slowly(asyncyield, count_to=count_to):
        ...     for i in range(count_to):
        ...         yield from asyncio.sleep(1)
        ...         yield from asyncyield(i + 1)
        >>> counter = count_slowly(10)
        >>> while (yield from counter.can_peek()):
        ...     i = counter.get_nowait()
        ...     print("Current count is %d"%i)
        """

        def decorate(function):
            cofun = asyncio.coroutine(function)
            def wrapped(*args, **kwargs):
                result = cls(maxsize=maxsize)
                def guarding():
                    running = cofun(result.put, *args, **kwargs)
                    try:
                        yield from running
                    except Exception as e:
                        yield from result.put_exception(e)
                    else:
                        yield from result.finish()
                asyncio.Task(guarding())
                return result
            return wrapped
        return decorate

    @classmethod
    def merge(cls, queues):
        """Asyncio's `as_completed` does not work with QueueWithEnd objects for
        the same reason it can't replace it (missing end-of-loop indication);
        the `merge` classmethod can be used instead to fetch results
        indiscriminately from queues as they are completed:

        >>> @QueueWithEnd.cogenerator()
        >>> def count(asyncyield):
        ...     for i in range(3):
        ...         yield from asyncyield(i + 1)
        ...         yield from time.sleep(0.1 * i)
        >>> firstcount = count()
        >>> secondcount = count()
        >>> merged = QueueWithEnd.merged([firstcount, secondcount])
        >>> while (yield from merged.can_peek()):
        ...     print(merged.get_nowait())
        1
        2
        1
        2
        3
        3
        """
        merged = cls(maxsize=1)
        merged.subqueues = queues[:]

        @asyncio.coroutine
        def feeder(queue, merged):
            while (yield from queue.can_peek()):
                if queue._flag == cls.Type.end:
                    merged.subqueues.remove(queue)
                    if not merged.subqueues:
                        merged.finish()
                        return
                yield from merged._put(queue._flag, queue._value)
                queue._flag = cls.Type.notpeeked
        for s in merged.subqueues:
            asyncio.Task(feeder(s, merged))

        return merged

    # implementing the Future interface -- note that it's neither a Future by
    # inheritance, nor does it offer the complete Future interface; but it can
    # be used in `for value in (yield from ...):`

    def __iter__(self):
        result = []
        while (yield from self.can_peek()):
            result.append(self.get_nowait())
        return result

    # compatibility to the original `Its` class

    more = can_peek
    value = property(get_nowait)
    # another old name
    consume = get_nowait



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""Tools not directly related with CoAP that are needed to provide the API"""


class ExtensibleEnumMeta(type):
    """Metaclass for ExtensibleIntEnum, see there for detailed explanations"""

    def __init__(self, name, bases, dict):
        self._value2member_map_ = {}
        for k, v in dict.items():
            if k.startswith('_'):
                continue
            if callable(v):
                continue
            if isinstance(v, property):
                continue
            instance = self(v)
            instance.name = k
            setattr(self, k, instance)
        type.__init__(self, name, bases, dict)

    def __call__(self, value):
        if isinstance(value, self):
            return value
        if value not in self._value2member_map_:
            self._value2member_map_[value] = super(ExtensibleEnumMeta, self).__call__(value)
        return self._value2member_map_[value]


class ExtensibleIntEnum(int, metaclass=ExtensibleEnumMeta):
    """Similar to Python3.4's enum.IntEnum, this type can be used for named
    numbers which are not comprehensively known, like CoAP option numbers."""

    def __add__(self, delta):
        return type(self)(int(self) + delta)

    def __repr__(self):
        return '<%s %d%s>' % (type(self).__name__, self, ' "%s"' % self.name if hasattr(self, "name") else "")

    def __str__(self):
        return self.name if hasattr(self, "name") else int.__str__(self)


def hostportjoin(host, port=None):
    """Join a host and optionally port into a hostinfo-style host:port
    string"""
    if ':' in host:
        host = '[%s]' % host

    if port is None:
        hostinfo = host
    else:
        hostinfo = "%s:%d" % (host, port)
    return hostinfo



import unittest
from Bubot.Helpers.JsonSchema.JsonSchema4 import JsonSchema4


class TestLoadSchema(unittest.TestCase):

    def test_load_schema(self):
        schema = JsonSchema4()

    def test_load_bubot_schema(self):
        dir = ['D:\\project\\bubot3\\bubot_AdminPanel\\src\\BubotObj\\OcfSchema\\schema', 'D:\\project\\bubot3\\bubot_Core\\src\\BubotObj\\OcfSchema\\schema', 'D:\\project\\bubot3\\bubot_Modbus\\src\\BubotObj\\OcfSchema\\schema', 'D:\\project\\bubot3\\bubot_PidController\\src\\BubotObj\\OcfSchema\\schema', 'D:\\project\\bubot3\\bubot_WebServer\\src\\BubotObj\\OcfSchema\\schema']
        rt = ['oic.r.openlevel']
        json_schema = JsonSchema4(dir=dir)
        # try:
        res = json_schema.load_from_rt(rt)
        # except Exception as err:
        #     a = 1
        # pass.

    def test_load_bubot_schema2(self):
        dir = ['D:\\project\\bubot3\\bubot_AdminPanel\\src\\BubotObj\\OcfSchema\\schema', 'D:\\project\\bubot3\\bubot_Core\\src\\BubotObj\\OcfSchema\\schema', 'D:\\project\\bubot3\\bubot_Modbus\\src\\BubotObj\\OcfSchema\\schema', 'D:\\project\\bubot3\\bubot_PidController\\src\\BubotObj\\OcfSchema\\schema', 'D:\\project\\bubot3\\bubot_WebServer\\src\\BubotObj\\OcfSchema\\schema']
        rt = ['oic.r.switch.binary']
        rt = ['bubot.serialserver.con', 'oic.wk.con', 'bubot.con']
        json_schema = JsonSchema4(dir=dir)
        # try:
        res = json_schema.load_from_rt(rt)
        # except Exception as err:
        #     a = 1
        pass


import socket
import asyncio
import struct
from Bubot.Helpers.Coap.CoapProtocol import CoapProtocol
from Bubot.Helpers.ExtException import ExtTimeoutError, CancelOperation


class CoapServer:
    coap_discovery_ipv6 = ['FF02::158', 'FF03::158', 'FF05::158']
    coap_discovery_ipv4 = ['224.0.1.187']  # All CoAP Nodes RFC7252
    coap_timeout = 25

    def __init__(self, device):
        self.device = device
        self._mid = 0
        self._token = 0
        self.log = device.log
        self.loop = device.loop
        self.ipv6 = self.device.get_param('/oic/con', 'udpCoapIPv6', True)
        self.ipv4 = self.device.get_param('/oic/con', 'udpCoapIPv4', False)
        self.multicast_port = 5683
        self.unicast_port = self.device.get_param('/oic/con', 'udpCoapPort', None)
        self.net_interface = self.device.get_param('/oic/con', 'NetInterface', 0)  # default interface

        self.resources = {}
        self.waiting_response = {}
        self.endpoint = {
            'multicast': [],
            'IPv4': {},
            'IPv6': {}
        }
        self._last_message_id = 0
        self.answer = {}

    @property
    def token(self):
        self._token = (self._token + 1) & 0xff
        return self._token

    @property
    def mid(self):
        self._mid = (self._mid + 1) & 0xffff
        return self._mid

    async def run(self):
        if self.ipv6:
            await self.run_multicast_ipv6(CoapProtocol, self.coap_discovery_ipv6, self.multicast_port)
            await self.run_unicast_ipv6(CoapProtocol, self.unicast_port)
        if self.ipv4:
            await self.run_multicast_ipv4(CoapProtocol, self.coap_discovery_ipv4, self.multicast_port)
            await self.run_unicast_ipv4(CoapProtocol, self.unicast_port)
        pass
        if self.device.get_param('/oic/con', 'udpCoapPort', None) != self.unicast_port:
            self.device.set_param('/oic/con', 'udpCoapPort', self.unicast_port, save_config=True)
        self.log.debug('on port {0}'.format(self.unicast_port))

    async def run_multicast_ipv6(self, protocol_factory, address, port):
        try:
            interface_index = 0  # default

            sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            # sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            # sock.setsockopt(41, socket.IPV6_V6ONLY, 0)
            sock.bind(('::', port))
            # sock.bind(('::', port))
            for group in address:
                sock.setsockopt(
                    41,  # socket.IPPROTO_IPV6 = 41 - not found in windows 10, bug python
                    socket.IPV6_JOIN_GROUP,
                    struct.pack(
                        '16si',
                        socket.inet_pton(socket.AF_INET6, group),
                        interface_index
                    )
                )
            sock.setsockopt(41, socket.IPV6_MULTICAST_LOOP, 1)
            _transport, _protocol = await self.loop.create_datagram_endpoint(
                lambda: protocol_factory(self, multicast=True),
                sock=sock,
            )
            # _address = _transport.get_extra_info('socket').getsockname()
            self.endpoint['multicast'].append(
                dict(
                    transport=_transport,
                    protocol=_protocol,
                    # uri='coap://[{0}]:{1}'.format(_address[0], _address[1])
                )
            )
        except Exception as e:
            self.log.error(e)
            raise e from None

    async def run_multicast_ipv4(self, protocol_factory, address, port):
        # interface_index = 0  # default
        # addrinfo = socket.getaddrinfo(address[0], None)[0]
        # sock = socket.socket(addrinfo[0], socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        # group_bin = socket.inet_pton(addrinfo[0], addrinfo[4][0])
        sock.bind(('', port))
        # mreq = group_bin + struct.pack('=I', socket.INADDR_ANY)
        # sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        for group in address:
            sock.setsockopt(
                socket.IPPROTO_IP,
                socket.IP_ADD_MEMBERSHIP,
                struct.pack(
                    "=4sl",
                    socket.inet_aton(group),
                    socket.INADDR_ANY
                )
            )
        _transport, _protocol = await self.loop.create_datagram_endpoint(
            lambda: protocol_factory(self, multicast=True),
            sock=sock,
        )
        self.endpoint['multicast'].append(
            dict(
                transport=_transport,
                protocol=_protocol,
                # uri='coap://[{0}]:{1}'.format(_address[0], _address[1])
            )
        )

    async def run_unicast_ipv6(self, protocol_factory, port=None):
        try:
            interface_index = 0  # default
            sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            # sock.setsockopt(41, socket.IPV6_V6ONLY, 0)
            sock.bind(('::', port))
            _transport, _protocol = await self.loop.create_datagram_endpoint(
                lambda: protocol_factory(self, multicast=True),
                sock=sock,
            )

            _address = _transport.get_extra_info('socket').getsockname()
            if self.unicast_port and self.unicast_port != _address[1]:
                raise Exception('IPv6 unicast port {} not installed'.format(self.unicast_port))
            _address = socket.getaddrinfo(socket.gethostname(), _address[1], socket.AF_INET6, socket.SOCK_DGRAM)[0][4]

            self.unicast_port = _address[1]
            self.endpoint['IPv6'] = dict(
                transport=_transport,
                protocol=_protocol,
                address=_address,
                uri='coap://[{0}]:{1}'.format(_address[0], _address[1])
            )
            pass
        except Exception as e:
            self.log.error(e)
            raise e from None

    async def run_unicast_ipv4(self, protocol_factory, port=None):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.bind(('', port))
        _transport, _protocol = await self.loop.create_datagram_endpoint(
            lambda: protocol_factory(self),
            sock=sock
        )
        _address = _transport.get_extra_info('socket').getsockname()
        if self.unicast_port and self.unicast_port != _address[1]:
            raise Exception('IPv4 unicast port {} not installed'.format(self.unicast_port))
        _address = socket.getaddrinfo(socket.gethostname(), _address[1], socket.AF_INET, socket.SOCK_DGRAM)[0][4]
        self.unicast_port = _address[1]
        self.endpoint['IPv4'] = dict(
            transport=_transport,
            protocol=_protocol,
            address=_address,
            uri='coap://{0}:{1}'.format(_address[0], _address[1])
        )

    async def send_multi_answer_request(self, message, address, callback, *args):
        self.log.debug('send_multi_request {0}'.format(message.opt.uri_path))
        self.answer[message.token] = dict(
            request=message,
            response=callback,
            lock=asyncio.Lock(),
            result=args[0] if args else None
        )
        self.send_message(message, address)

    async def send_answer(self, answer):
        msg, remote = answer.encode_to_coap()
        self.send_message(msg, remote)

    async def send_request(self, message, remote, **kwargs):
        _request_description = f'{remote[0]}:{remote[1]}{message.opt.uri_path}'
        self.log.debug(_request_description)
        self.answer[message.token] = dict(
            request=message,
            response=asyncio.Future()
        )

        self.send_message(message, remote)

        timeout = kwargs.get('timeout', self.coap_timeout)
        try:
            res = await asyncio.wait_for(self.answer[message.token]['response'], timeout)
            self.log.debug('receiv_answer {0} {1}'.format(
                message.token, remote[0]))
            return res
        except KeyError:
            self.log.error('unknown')
            raise
        except asyncio.CancelledError:
            raise CancelOperation(detail=_request_description)
        except asyncio.TimeoutError:
            raise ExtTimeoutError(detail=_request_description)
        finally:
            self.answer.pop(message.token)
            pass

    def send_message(self, message, remote):
        try:
            net_index = remote[0].find('%')
        except AttributeError:
            raise ValueError('Receiver address  not defined')
        try:
            if net_index >= 0:
                remote = (remote[0][0:net_index], remote[1])  # убираем номер интерфейса из адреса
            _protocol = 'IPv6' if socket.getaddrinfo(remote[0], remote[1])[0][0] == socket.AF_INET6 else 'IPv4'
            try:
                transport = self.endpoint[_protocol]['transport']
            except KeyError:
                raise Exception(f'protocol not supported')
            raw_data = message.encode()
            self.log.debug(f'send message {message.mid} {remote}')
            transport.sendto(raw_data, remote)
            pass
        except Exception as e:
            self.log.error(e)
            raise e from None

    def close(self):
        if self.ipv6:
            self.endpoint['IPv6']['transport'].close()
        if self.ipv4:
            self.endpoint['IPv4']['transport'].close()
        for elem in self.endpoint['multicast']:
            elem['transport'].close()



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""This module contains numeric constants that would be expected in the socket
module, but are not exposed there.

For some platforms (eg. python up to 3.5 on Linux), there is an IN module that
exposes them; and they are gathered from there.

As a fallback, the numbers are hardcoded. Any hints on where to get them from
are appreciated; possible options are parsing C header files (at build time?)
or interacting with shared libraries for obtaining the symbols. The right way
would probably be including them in Python.
"""

try:
    from IN import IPV6_RECVERR, IP_RECVERR, IPV6_PKTINFO
except ImportError:
    IPV6_RECVERR = 25
    IP_RECVERR = 11
    IPV6_PKTINFO = 50



import asyncio
import inspect
from .ExtException import ExtException, HandlerNotFoundError

import copy
from collections import OrderedDict
from datetime import datetime, timedelta
import json
import os
import re


class Helper:

    @staticmethod
    def get_obj_class(path, obj_name, **kwargs):
        folder_name = f'.{obj_name}' if kwargs.get('folder') else ''
        class_name = obj_name
        suffix = kwargs.get('suffix')
        if suffix:
            class_name += suffix
        full_path = f'{path}{folder_name}.{class_name}.{class_name}'
        try:
            return Helper.get_class(full_path)
        except ExtException as err:
            raise HandlerNotFoundError(detail=f'object {obj_name}', parent=err)

    @staticmethod
    def get_class(class_full_path):
        try:
            parts = class_full_path.split('.')
            module = ".".join(parts[:-1])
            m = __import__(module)
            for comp in parts[1:]:
                m = getattr(m, comp)
            return m
        except ImportError as err:
            # ошибки в классе  или нет файла
            raise ExtException(
                'Unable to import object',
                detail=class_full_path,
                action='Helper.get_class',
                parent=err

            )
        except AttributeError as err:
            # Нет такого класса
            # raise AttributeError('get_class({0}: {1})'.format(class_full_path, str(e)))
            raise ExtException(
                'Object not found',
                detail=class_full_path,
                action='Helper.get_class'
            )
        except Exception as err:
            raise ExtException(
                'Unable to import object',
                detail=class_full_path,
                action='Helper.get_class'
            )

    @classmethod
    def loads_exception(cls, data):
        try:
            data = json.loads(data)
            _handler = Helper.get_class(f'{data["__module__"]}.{data["__name__"]}')
            return _handler(**data)
        except ExtException as err:
            return ExtException(data, action='Helper.loads_exception', parent=err)

    @staticmethod
    def convert_ticks_to_datetime(ticks):
        return datetime(1, 1, 1) + timedelta(microseconds=ticks // 10)

    @staticmethod
    def update_dict(*args):
        size = len(args)
        if size == 1:
            return Helper._update_dict({}, args[0])
        elif size > 1:
            result = args[0]
            for i in range(size - 1):
                result = Helper._update_dict(result, args[i + 1])
            return result

    @staticmethod
    def _update_dict(base, new):
        if not new:
            return base
        for element in new:
            try:
                if element in base and base[element] is not None:
                    if isinstance(new[element], dict):
                        base[element] = Helper._update_dict(base[element], new[element])
                    elif isinstance(new[element], list):
                        base[element] = ArrayHelper.unique_extend(base[element], new[element])
                    else:
                        base[element] = new[element]
                else:
                    try:
                        base[element] = copy.deepcopy(new[element])
                    except TypeError as e:
                        if not base:
                            base = {
                                element: copy.deepcopy(new[element])
                            }
                        else:
                            raise Exception('Непредвиденная ситуация при ')
            except ExtException as e:
                try:
                    _elem = '{0}.{1}'.format(element, e.stack[len(e.stack) - 1]['dump']['element'])
                    _msg = e.stack[0]['dump']['message']
                except ExtException:
                    _msg = str(e)
                    _elem = element
                raise ExtException(
                    e,
                    action='Helper.update_dict',
                    detail='{1} - {0}'.format(_msg, _elem),
                    dump={
                        'element': _elem})
            except Exception as e:
                raise ExtException(
                    605,
                    action='Helper.update_dict',
                    detail='{0}({1})'.format(e, element),
                    dump={
                        'element': element,
                        'message': str(e)
                    })
        return base

    @classmethod
    def xml_to_json(cls, elem, array_mode):
        if not array_mode:
            d = OrderedDict()
            for key, value in elem.attrib.items():
                # d.__setitem__(d, "key", value)
                d[key] = value
            elem_text = elem.text
            if elem_text is None:
                d = elem.tag
            else:
                elem_text = elem_text.strip()
                if elem_text:
                    # d.__setitem__(d, "Значение", elem_text)
                    d["Значение"] = elem_text
        else:
            d = []
        array = OrderedDict()
        names_check = {"main": [], "sub": {}}
        for sub_elem in elem:
            this_array = True if sub_elem.find("[@Имя]") else False
            try:
                value = cls.xml_to_json(sub_elem, this_array)
            except Exception as error:
                element = sub_elem.tag + (
                    "." + sub_elem.get('Имя') if this_array else "") + "." + error.detail if hasattr(
                    error, 'detail') else ''
                raise ExtException(error, dump={'element': element})
            if this_array:
                item_name = sub_elem.get('Имя')
                if array_mode:
                    # Проверяем повторение подэлементов при наличии узла Имя, так как все одноименные узлы
                    # с таким атрибутом будут слиты в 1, а по атрибуту "Имя" добавлены как элементы.
                    # Одинаковых атрибутов "Имя" в рамках одноименных узлов быть не должно
                    sub_checked = []
                    if sub_elem.tag in names_check["sub"]:
                        sub_checked = names_check["sub"][sub_elem.tag]
                    if item_name in sub_checked:
                        raise ExtException('Дублирующиеся узлы в xml файле.', dump=sub_elem.tag + "." + item_name)
                    sub_checked.append(item_name)
                    names_check["sub"][sub_elem.tag] = sub_checked
                    if sub_elem.tag not in array:
                        if sub_elem.tag == item_name:
                            array[sub_elem.tag] = value
                        else:
                            array[sub_elem.tag] = OrderedDict()
                        d.append({"Имя": sub_elem.tag, "Значение": array[sub_elem.tag]})
                    if sub_elem.tag != item_name:
                        array[sub_elem.tag][item_name] = value
                else:
                    if sub_elem.tag not in d:
                        d[sub_elem.tag] = OrderedDict()
                    if sub_elem.tag == item_name:
                        d[sub_elem.tag] = value
                    else:
                        d[sub_elem.tag][item_name] = value
            else:
                if array_mode:
                    # Тут просто проверям наличие одноименных узлов без атрибута "Имя".
                    if sub_elem.tag in names_check["main"]:
                        raise ExtException('Дублирующиеся узлы в xml файле.', dump=sub_elem.tag + "." + item_name)
                    names_check["main"].append(sub_elem.tag)
                    if isinstance(value, str):
                        d.append(value)
                    else:
                        value_new = OrderedDict()  # Для порядка Имя-Значение в результате делаем новый результат.
                        value_new["Имя"] = sub_elem.tag
                        value_new.update(value)
                        # for value_key in value.keys():
                        #     value_new[value_key] = value[value_key]
                        d.append(value_new)
                else:
                    d[sub_elem.tag] = value
        return d

    @classmethod
    def update_element_in_dict(cls, _data, _path, _value):
        _num = re.compile('^\d+$')
        _path = _path.split('.')
        current = _data
        size = len(_path)
        for i, elem in enumerate(_path):
            if _num.match(elem):
                elem = int(elem)
            if i == size - 1:
                current[elem] = _value
            else:
                current = current[elem]

    @classmethod
    def compare(cls, base, new):
        def compare_dict(_base, _new):
            pass

        def compare_list(_base, _new):
            pass

        if isinstance(base, dict):
            difference = False
            res = {}
            for elem in new:
                try:
                    if base and elem in base:
                        if isinstance(new[elem], dict):
                            _difference, _res = cls.compare(base[elem], new[elem])
                            if _difference:
                                difference = True
                                res[elem] = copy.deepcopy(_res)
                        else:
                            if new[elem] != base[elem]:
                                difference = True
                                res[elem] = new[elem]
                    else:
                        difference = True
                        res[elem] = new[elem]
                except Exception as e:
                    raise Exception('compare: {0}'.format(str(e)), elem)
        else:
            difference = False
            res = None
            if base != new:
                difference = True
                res = new
        return difference, res

    @classmethod
    def get_default_config(cls, current_class, root_class, cache):
        _type = root_class.__name__
        try:
            return cache['{0}_{1}'.format(_type, cls.__name__)]
        except KeyError:
            pass
        schema = {}

        for elem in current_class.__bases__:
            if issubclass(elem, root_class):
                try:
                    _schema = cache['{0}_{1}'.format(_type, elem.__name__)]
                except KeyError:
                    _schema = cls.get_default_config(elem, root_class, cache)
                    cache['res' + elem.__name__] = Helper.update_dict({}, _schema)
                Helper.update_dict(schema, _schema)
        if hasattr(current_class, 'file'):
            config_path = '{0}/{1}.json'.format(os.path.dirname(current_class.file), current_class.__name__)
            try:
                with open(config_path, 'r', encoding='utf-8') as file:
                    _schema = json.load(file)
                    Helper.update_dict(schema, _schema)
            except FileNotFoundError:
                return schema
            except Exception as e:
                pass
        cache['{0}_{1}'.format(_type, current_class.__name__)] = Helper.update_dict({}, schema)
        return schema


def async_test(f):
    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        # kwargs['loop'] = loop
        if inspect.iscoroutinefunction(f):
            future = f(*args, **kwargs)
        else:
            coroutine = asyncio.coroutine(f)
            future = coroutine(*args, **kwargs)
        loop.run_until_complete(future)

    return wrapper


def convert_ticks_to_datetime(ticks):
    return datetime(1, 1, 1) + timedelta(microseconds=ticks // 10)


class ArrayHelper:
    # делаем список уникальных записей, объединяемые массивы должны быть однотипными
    # списки объектов объединаются по ключевому полю которое есть в первой записи списка
    default_object_uid_props = ['_id', 'id', 'di', 'title', 'name', 'n']

    @classmethod
    def unique_extend(cls, a, b, **kwargs):
        if not len(a):
            a.extend(b)
            return a
        if not len(b):
            return a
        if isinstance(a[0], str):
            for elem in b:
                if elem not in a:
                    a.append(elem)
            return a
        if isinstance(a[0], dict):
            object_uid_prop = cls.detect_object_uid_prop(a, kwargs.get('object_uid', cls.default_object_uid_props))
            if object_uid_prop is None:
                c = copy.deepcopy(b)
                a.extend(c)
                return a
            index = cls.index_list(a, object_uid_prop)
            for elem in b:
                uid = elem.get(object_uid_prop)
                if uid not in index:
                    a.append(elem)
        return a


    @classmethod
    def detect_object_uid_prop(cls, data, object_uid_fields):
        for _key in object_uid_fields:
            if _key in data[0]:
                return _key
        return None

    @classmethod
    def update(cls, items, item, field='id'):
        for i in range(len(items)):
            if items[i][field] == item[field]:
                items[i] = item
                return
        items.append(item)

    @classmethod
    def index_list(cls, items, field='id'):
        res = {}
        for index in range(len(items)):
            try:
                res[items[index][field]] = index
            except KeyError:
                pass
        return res

    @classmethod
    def find(cls, items, value, field='id'):
        for i in range(len(items)):
            if items[i][field] == value:
                return i
        return -1



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""
Exception definitions for txThings CoAP library.
"""

import warnings
import abc

from .numbers import codes


class Error(Exception):
    """
    Base exception for all exceptions that indicate a failed request
    """


class RenderableError(Error, metaclass=abc.ABCMeta):
    """
    Exception that can meaningfully be represented in a CoAP response
    """

    @abc.abstractmethod
    def to_message(self):
        """Create a CoAP message that should be sent when this exception is
        rendered"""


class ResponseWrappingError(Error):
    """
    An exception that is raised due to an unsuccessful but received response.

    A better relationship with :mod:`.numbers.codes` should be worked out to do
    `except UnsupportedMediaType`.
    """

    def __init__(self, coapmessage):
        self.coapmessage = coapmessage

    def to_message(self):
        return self.coapmessage

    def __repr__(self):
        return "<%s: %s %r>" % (type(self).__name__, self.coapmessage.code, self.coapmessage.payload)


class ConstructionRenderableError(RenderableError):
    """
    RenderableError that is constructed from class attrinutes :attr:`code` and
    :attr:`message`
    """
    code = codes.INTERNAL_SERVER_ERROR
    message = ""


# FIXME: this should be comprehensive, maybe generted from the code list

class NotFound(ConstructionRenderableError):
    code = codes.NOT_FOUND


class MethodNotAllowed(ConstructionRenderableError):
    code = codes.METHOD_NOT_ALLOWED


class UnsupportedMediaType(ConstructionRenderableError):
    code = codes.UNSUPPORTED_MEDIA_TYPE


class BadRequest(ConstructionRenderableError):
    code = codes.BAD_REQUEST


# More detailed versions of code based errors

class NoResource(NotFound):
    """
    Raised when resource is not found.
    """
    message = "Error: Resource not found!"

    def __init__(self):
        warnings.warn("NoResource is deprecated in favor of NotFound", DeprecationWarning, stacklevel=2)


class UnallowedMethod(MethodNotAllowed):
    """
    Raised by a resource when request method is understood by the server
    but not allowed for that particular resource.
    """
    message = "Error: Method not allowed!"


class UnsupportedMethod(MethodNotAllowed):
    """
    Raised when request method is not understood by the server at all.
    """
    message = "Error: Method not recognized!"


class NotImplemented(Error):
    """
    Raised when request is correct, but feature is not implemented
    by txThings library.
    For example non-sequential blockwise transfers
    """


class RequestTimedOut(Error):
    """
    Raised when request is timed out.
    """


class WaitingForClientTimedOut(Error):
    """
    Raised when server expects some client action:

        - sending next PUT/POST request with block1 or block2 option
        - sending next GET request with block2 option

    but client does nothing.
    """


class ResourceChanged(Error):
    """
    The requested resource was modified during the request and could therefore
    not be received in a consistent state.
    """


class UnexpectedBlock1Option(Error):
    """
    Raised when a server responds with block1 options that just don't match.
    """


class UnexpectedBlock2(Error):
    """
    Raised when a server responds with another block2 than expected.
    """


class MissingBlock2Option(Error):
    """
    Raised when response with Block2 option is expected
    (previous response had Block2 option with More flag set),
    but response without Block2 option is received.
    """


class NotObservable(Error):
    """
    The server did not accept the request to observe the resource.
    """


class ObservationCancelled(Error):
    """
    The server claimed that it will no longer sustain the observation.
    """


class UnparsableMessage(Error):
    """
    An incoming message does not look like CoAP.

    Note that this happens rarely -- the requirements are just two bit at the
    beginning of the message, and a minimum length.
    """


class CommunicationKilled(ConstructionRenderableError):
    """
    The communication process has been aborted by request of the application.
    """
    code = codes.SERVICE_UNAVAILABLE



import json
import urllib.parse
from Bubot.Helpers.Helper import Helper
from Bubot.Helpers.ExtException import ExtException
from os import path
import os.path


class JsonSchemaLoaderMixin:
    def __init__(self, **kwargs):
        self.dir = kwargs.get('dir', ['{}/OcfSchema'.format(path.normpath(path.dirname(__file__)))])
        self.cache = kwargs.get('cache', {})
        self.data = {}
        self.id = None
        self.uri = None
        self.version = None
        self.load_history = []

    def load_from_uri(self, uri, **kwargs):
        try:
            uri = urllib.parse.urlparse(uri)  # (scheme, netloc, path, params, query, fragment)
            _id = self.get_id_from_uri(uri, self.id)

            if _id not in self.cache:
                JsonSchema4.load_from_file(_id, dir=self.dir, cache=self.cache)
            if not uri[5]:
                try:
                    return self.cache[_id]['data']
                except KeyError:
                    raise Exception('bad schema id or definition in {}'.format(_id))

            _fragment = uri[5].split('/')[1:]
            _result = self.cache[_id]
            for elem in _fragment:
                _result = _result[elem]
            return _result
        except Exception as err:
            raise ExtException(
                action='JsonSchemaLoaderMixin.load_from_uri',
                dump={'uri': uri},
                parent=err) from err

    @classmethod
    def load_from_file(cls, file_name, **kwargs):
        try:
            self = cls(**kwargs)
            _path = None
            for _dir in self.dir:
                _path = '{0}/{1}'.format(_dir, file_name)
                if os.path.isfile(_path):
                    break
                else:
                    _path = None
            if _path is None:
                raise ExtException(message='Json schema file not found', detail=file_name)

            with open(_path, 'r', encoding='utf-8') as file:
                try:
                    raw = json.load(file)
                except Exception as err:
                    raise ExtException(message='JsonSchema not load from file', detail=f'{file_name} {str(err)}') from err
            self.load(raw)
            return self.cache[self.id]['data']
        except Exception as err:
            raise ExtException(
                action='JsonSchemaLoaderMixin.load_from_file',
                dump={'file_name': file_name},
                parent=err) from err

    def load_from_rt(self, rt):
        schema = {}
        for name in rt:
            uri = '{0}-schema.json'.format(name)
            _schema = self.load_from_uri(uri)
            Helper.update_dict(schema, _schema)
        return schema

    @staticmethod
    def get_id_from_uri(uri, default=None):
        if uri[2]:
            return uri[2].split('/')[-1]
        elif default:
            return default
        raise Exception('not define scheme id')

    def load(self, schema):
        try:
            self.version = urllib.parse.urlparse(schema['$schema'])[2].split('/')[0]
            self.id = self.get_id_from_uri(urllib.parse.urlparse(schema['id']))
            self.cache[self.id] = dict(data=dict(), definitions=dict())
            if 'definitions' in schema:
                for tmpl in schema['definitions']:
                    self.cache[self.id]['definitions'][tmpl] = dict()
                    for props in schema['definitions'][tmpl]:
                        _type = schema['definitions'][tmpl][props].__class__.__name__
                        self.load_elem(
                            self.cache[self.id]['definitions'][tmpl],
                            schema['definitions'][tmpl][props],
                            props
                        )
            for elem in schema:
                if elem != 'definitions':
                    self.load_elem(self.cache[self.id]['data'], schema[elem], elem)
            pass
        except Exception as err:
            raise ExtException(
                action='JsonSchemaLoaderMixin.load',
                parent=err) from err

    # @staticmethod
    # def add_prop(data, node, name, default=None):
    #     try:
    #         data[name] = node['name']
    #     except KeyError:
    #         if default is not None:
    #             data[name] = default

    def load_elem_ref(self, data, value, name):
        res = self.load_from_uri(value)
        Helper.update_dict(data, res)

    def load_elem(self, data, value, name):
        try:
            value_type = value.__class__.__name__
            getattr(self, f'load_elem_{value_type}')(
                data, value, name)
        except Exception as err:
            raise ExtException(
                action='JsonSchemaLoaderMixin.load_elem',
                dump={'elem_name': name},
                parent=err) from err
        pass

    def load_elem_str(self, data, value, name):
        if name != '$ref':
            data[name] = value
            return
        self.load_elem_ref(data, value, name)

    def load_elem_dict(self, data, value, name):
        if name not in data:
            data[name] = dict()
        for elem in value:
            self.load_elem(data[name], value[elem], elem)

    def load_elem_list(self, data, value, name):
        if name == 'allOf':
            for elem in value:
                for _name in elem:
                    self.load_elem(data, elem[_name], _name)
            return
        else:
            if name not in data:
                data[name] = value
            else:
                data[name] = list(set(data[name]+ value))
        pass

    def load_elem_bool(self, data, value, name):
        data[name] = value

    def load_elem_int(self, data, value, name):
        data[name] = value

    def load_elem_float(self, data, value, name):
        data[name] = value


class JsonSchema4(JsonSchemaLoaderMixin):
    def __init__(self, **kwargs):
        JsonSchemaLoaderMixin.__init__(self, **kwargs)



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""asyncio workarounds"""

import asyncio.events


def cancel_thoroughly(handle):
    """Use this on a (Timer)Handle when you would .cancel() it, just also drop
    the callback and arguments for them to be freed soon."""

    assert isinstance(handle, asyncio.events.Handle)

    handle.cancel()
    handle._args = handle._callback = None


from asyncio import DatagramProtocol
from asyncio.selector_events import _SelectorDatagramTransport, BaseSelectorEventLoop
import socket


class RecvmsgDatagramProtocol(DatagramProtocol):
    """Inheriting from this indicates that the instance expects to be called
    back datagram_msg_received instead of datagram_received"""


class RecvmsgSelectorDatagramTransport(_SelectorDatagramTransport):
    def __init__(self, *args, **kwargs):
        super(RecvmsgSelectorDatagramTransport, self).__init__(*args, **kwargs)

    def _read_ready(self):
        try:
            data, ancdata, flags, addr = self._sock.recvmsg(self.max_size, 1024, socket.MSG_ERRQUEUE)
        except (BlockingIOError, InterruptedError):
            pass
        except OSError as exc:
            self._protocol.error_received(exc)
        except Exception as exc:
            self._fatal_error(exc, 'Fatal read error on datagram transport')
        else:
            self._protocol.datagram_errqueue_received(data, ancdata, flags, addr)

        # copied and modified from _SelectorDatagramTransport
        try:
            data, ancdata, flags, addr = self._sock.recvmsg(self.max_size,
                                                            1024)  # TODO: find a way for the application to tell the trensport how much data is expected
        except (BlockingIOError, InterruptedError):
            pass
        except OSError as exc:
            self._protocol.error_received(exc)
        except Exception as exc:
            self._fatal_error(exc, 'Fatal read error on datagram transport')
        else:
            self._protocol.datagram_msg_received(data, ancdata, flags, addr)

    def sendmsg(self, data, ancdata, flags, address):
        # copied and modified from _SelectorDatagramTransport.sendto
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError('data argument must be a bytes-like object, '
                            'not %r' % type(data).__name__)
        if not data:
            return

        if self._address and addr not in (None, self._address):
            raise ValueError('Invalid address: must be None or %s' %
                             (self._address,))

        if self._conn_lost and self._address:
            if self._conn_lost >= constants.LOG_THRESHOLD_FOR_CONNLOST_WRITES:
                logger.warning('socket.send() raised exception.')
            self._conn_lost += 1
            return

        if not self._buffer:
            # Attempt to send it right away first.
            try:
                self._sock.sendmsg((data,), ancdata, flags, address)
                return
            except (BlockingIOError, InterruptedError):
                self._loop.add_writer(self._sock_fd, self._sendto_ready)
            except OSError as exc:
                self._protocol.error_received(exc)
                return
            except Exception as exc:
                self._fatal_error(exc,
                                  'Fatal write error on datagram transport')
                return

        # Ensure that what we buffer is immutable.
        self._buffer.append((bytes(data), ancdata, flags, addr))
        self._maybe_pause_protocol()

    # TODO: not modified _sendto_ready as it's not used in this application and
    # would only be dead code -- given that we store 4-tuples instead of
    # 2-tuples, _sendto_ready will fail anyway cleanly


# monkey patching because otherwise we'd have to create a loop subclass and
# require this to be loaded.

_orig_mdt = BaseSelectorEventLoop._make_datagram_transport


def _new_mdt(self, sock, protocol, *args, **kwargs):
    if isinstance(protocol, RecvmsgDatagramProtocol):
        return RecvmsgSelectorDatagramTransport(self, sock, protocol, *args, **kwargs)
    else:
        return _orig_mdt(self, sock, protocol, *args, **kwargs)


BaseSelectorEventLoop._make_datagram_transport = _new_mdt



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""
aiocoap
=======

The aiocoap package is a library that implements CoAP, the Constrained
Application Protocol (`RFC 7252`_, more info at http://coap.technology/).

.. _`RFC 7252`: http://tools.ietf.org/html/rfc7252

Usage
-----

In all but the most exotic applications, you will want to create a single
:class:`.Context` instance that binds to the network. The
:meth:`.Context.create_client_context` and
:meth:`.Context.create_server_context` coroutines give you a readily connected
context.

On the client side, you can request resources by assembling a :class:`.Message`
and passing it to your context's :meth:`.Context.request` method, which
returns a :class:`.protocol.Request` object with a
:attr:`.protocol.Request.response` future (which is a :class:`.Message` again).

On the server side, a resource tree gets built from
:class:`aiocoap.resource.Resource` objects into a
:class:`aiocoap.resource.Site`, which is assigned to the context at creation
time.
"""

from .numbers import *
from .message import Message



from time import time
import json
from .ExtException import ExtException, CancelOperation
import asyncio


class Action:
    def __init__(self, name=None, begin=True):
        self.name = name
        self.param = {}
        # self.error = None
        self.result = None
        self.begin = None
        self.end = None
        self.time = 0
        self.total_time = 0
        self.stat = {}
        if begin:
            self.set_begin()

    def set_begin(self):
        self.begin = time()

    def set_end(self, result=None):
        if self.end:
            return self
        self.end = time()
        if not self.begin:
            self.begin = self.end
        self.total_time = round(self.end - self.begin, 3)
        if result is not None:
            self.result = result
        if self.name:
            self.update_stat(self.name, [self.total_time - self.time, 1])
        return self

    def add_stat(self, action):
        if not isinstance(action, Action):
            return action
        for elem in action.stat:
            self.update_stat(elem, action.stat[elem])
        return action.result

    def update_stat(self, name, stat):
        self.time += stat[0]
        if name not in self.stat:
            self.stat[name] = stat
        else:
            self.stat[name][1] += stat[1]
            self.stat[name][0] += stat[0]
        pass

    # def __bool__(self):
    #     return False if self.error else True

    # def __str__(self):
    #     pass

    def to_dict(self):
        return {
            'result': self.result,
            'stat': {
                'action': self.name,
                'time': self.total_time,
                'detail': self.stat
            }
        }

    def dump(self):
        return json.dumps(self.to_dict(), ensure_ascii=False)
        pass

    def load(self, json_string):
        _tmp = json.loads(json_string)
        self.name = _tmp.get('name')
        # self.error = _tmp.get('error', None)
        self.result = _tmp.get('result', None)
        self.begin = _tmp.get('begin', None)
        self.end = _tmp.get('end', None)
        self.time = _tmp.get('time', 0)
        self.stat = _tmp.get('stat', {})
        return self

    def ext_exception(self, error, **kwargs):
        kwargs['action'] = kwargs.get('action', self.name)
        kwargs['parent'] = error
        kwargs['skip_traceback'] = 1
        _class = kwargs.get('exception_class', ExtException)
        return _class(**kwargs)


def async_action(f):
    async def wrapper(*args, **kwargs):
        try:
            name = f'{args[0].__name__}.{f.__name__}'
        except Exception:
            try:
                name = f'{args[0].__class__.__name__}.{f.__name__}'
            except Exception:
                name = f.__name__
        action = Action(name)
        try:
            kwargs['_action'] = action
            result = await f(*args, **kwargs)
            result = action.add_stat(result)
            if not isinstance(result, Action):
                return result
            return action.set_end(result)
        except asyncio.CancelledError as err:
            raise action.ext_exception(err, action=name, exception_class=CancelOperation)
        except ExtException as err:
            err = action.ext_exception(err, action=name)
            raise err
        except Exception as err:
            raise action.ext_exception(err, action=name)

    return wrapper


def action(f):
    def wrapper(*args, **kwargs):
        try:
            name = f'{args[0].__name__}.{f.__name__}'
        except Exception:
            try:
                name = f'{args[0].__class__.__name__}.{f.__name__}'
            except Exception:
                name = f.__name__
        action = Action(name)
        try:
            kwargs['_action'] = action
            result = f(*args, **kwargs)
            result = action.add_stat(result)
            if not isinstance(result, Action):
                return result
            return action.set_end(result)
        except ExtException as err:
            raise action.ext_exception(err, action=name)
        except Exception as err:
            raise action.ext_exception(err, action=name)

    return wrapper



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

import urllib.parse
import struct
import copy

from . import error
from .numbers import *
from .options import Options
from .util import hostportjoin

## Monkey patch urllib to make URL joining available in CoAP
# This is a workaround for <http://bugs.python.org/issue23759>.
urllib.parse.uses_relative.append('coap')
urllib.parse.uses_netloc.append('coap')


class Message(object):
    """CoAP Message with some handling metadata

    This object's attributes provide access to the fields in a CoAP message and
    can be directly manipulated.

    * Some attributes are additional data that do not round-trip through
      serialization and deserialization. They are marked as "non-roundtrippable".
    * Some attributes that need to be filled for submission of the message can
      be left empty by most applications, and will be taken care of by the
      library. Those are marked as "managed".

    The attributes are:

    * :attr:`payload`: The payload (body) of the message as bytes.
    * :attr:`mtype`: Message type (CON, ACK etc, see :mod:`.numbers.types`).
      Managed unless set by the application.
    * :attr:`code`: The code (either request or response code), see
      :mod:`.numbers.codes`.
    * :attr:`opt`: A container for the options, see :class:`.options.Options`.

    * :attr:`mid`: The message ID. Managed by the :class:`.Context`.
    * :attr:`token`: The message's token as bytes. Managed by the :class:`.Context`.
    * :attr:`remote`: The socket address of the other side, managed by the
      :class:`.protocol.Request` by resolving the ``.opt.uri_host`` or
      ``unresolved_remote``, or the :class:`.Responder` by echoing the incoming
      request's. (If you choose to set this explicitly set this, make sure not
      to set incomplete IPv6 address tuples, as they can be sent but don't
      compare equally with the responses). Non-roundtrippable.

    * requested_*: Managed by the :class:`.protocol.Request` a response results
      from, and filled with the request's URL data. Non-roundtrippable.
    * unresolved_remote: ``host[:port]`` (strictly speaking; hostinfo as in a
      URI) formatted string. If this attribute is set, it overrides
      ``.opt.uri_host`` (and ``-_port``) when it comes to filling the
      ``remote`` in an outgoing request.

      Use this when you want to send a request with a host name that would not
      normally resolve to the destination address. (Typically, this is used for
      proxying.)

    * :attr:`prepath`, :attr:`postpath`: Not sure, will probably go away when
      resources are overhauled. Non-roundtrippable.

    Options can be given as further keyword arguments at message construction
    time. This feature is experimental, as future message parameters could
    collide with options.
    """

    def __init__(self, *, mtype=None, mid=None, code=None, payload=b'', token=None, uri=None, **kwargs):
        self.version = 1
        if mtype is None:
            # leave it unspecified for convenience, sending functions will know what to do
            self.mtype = None
        else:
            self.mtype = Type(mtype)
        self.mid = mid
        if code is None:
            # as above with mtype
            self.code = None
        else:
            self.code = Code(code)
        self.token = token
        self.payload = payload
        self.opt = Options()

        self.remote = None
        self.unresolved_remote = None
        self.prepath = None
        self.postpath = None

        # attributes that indicate which request path the response belongs to.
        # their main purpose is allowing .get_request_uri() to work smoothly, a
        # feature that is required to resolve links relative to the message.
        #
        # path and query are stored as lists, as they would be accessed for
        # example by self.opt.uri_path
        self.requested_proxy_uri = None
        self.requested_scheme = None
        self.requested_hostinfo = None
        self.requested_path = None
        self.requested_query = None

        # deprecation error, should go away roughly after 0.2 release
        if self.payload is None:
            raise TypeError("Payload must not be None. Use empty string instead.")

        if uri:
            self.set_request_uri(uri)

        for k, v in kwargs.items():
            setattr(self.opt, k, v)

    def __repr__(self):
        return "<aiocoap.Message at %#x: %s %s (ID %r, token %r) remote %s%s%s>" % (
            id(self),
            self.mtype,
            self.code,
            self.mid,
            self.token,
            self.remote,
            ", %s option(s)" % len(self.opt._options) if self.opt._options else "",
            ", %s byte(s) payload" % len(self.payload) if self.payload else ""
        )

    def copy(self, **kwargs):
        """Create a copy of the Message. kwargs are treated like the named
        arguments in the constructor, and update the copy."""
        # This is part of moving messages in an "immutable" direction; not
        # necessarily hard immutable. Let's see where this goes.
        new = copy.deepcopy(self)
        if 'mtype' in kwargs:
            new.mtype = Type(kwargs.pop('mtype'))
        if 'mid' in kwargs:
            new.mid = kwargs.pop('mid')
        if 'code' in kwargs:
            new.code = Code(kwargs.pop('code'))
        if 'payload' in kwargs:
            new.payload = kwargs.pop('payload')
        if 'token' in kwargs:
            new.token = kwargs.pop('token')
        if 'uri' in kwargs:
            new.set_request_uri(kwargs.pop('uri'))

        for k, v in kwargs.items():
            setattr(new.opt, k, v)

        return new

    @classmethod
    def decode(cls, rawdata, remote=None):
        """Create Message object from binary representation of message."""
        try:
            (vttkl, code, mid) = struct.unpack('!BBH', rawdata[:4])
        except struct.error:
            raise error.UnparsableMessage("Incoming message too short for CoAP")
        version = (vttkl & 0xC0) >> 6
        if version is not 1:
            raise error.UnparsableMessage("Fatal Error: Protocol Version must be 1")
        mtype = (vttkl & 0x30) >> 4
        token_length = (vttkl & 0x0F)
        msg = Message(mtype=mtype, mid=mid, code=code)
        msg.token = int.from_bytes(rawdata[4:4 + token_length], 'big')
        msg.payload = msg.opt.decode(rawdata[4 + token_length:])
        msg.remote = remote
        return msg

    def encode(self):
        """Create binary representation of message from Message object."""
        if self.code is None or self.mtype is None or self.mid is None:
            raise TypeError("Fatal Error: Code, Message Type and Message ID must not be None.")
        token = self.token.to_bytes(8, byteorder='big') if self.token else b''
        rawdata = bytes([(self.version << 6) + ((self.mtype & 0x03) << 4) + (len(token) & 0x0F)])
        rawdata += struct.pack('!BH', self.code, self.mid)
        rawdata += token
        rawdata += self.opt.encode()
        if len(self.payload) > 0:
            rawdata += bytes([0xFF])
            rawdata += self.payload
        return rawdata

    def get_cache_key(self, ignore_options=()):
        """Generate a hashable and comparable object (currently a tuple) from
        the message's code and all option values that are part of the cache key
        and not in the optional list of ignore_options (which is the list of
        option numbers that are not technically NoCacheKey but handled by the
        application using this method).

        >>> m1 = Message(code=GET)
        >>> m2 = Message(code=GET)
        >>> m1.opt.uri_path = ('s', '1')
        >>> m2.opt.uri_path = ('s', '1')
        >>> m1.opt.size1 = 10 # the only no-cache-key option in the base spec
        >>> m2.opt.size1 = 20
        >>> m1.get_cache_key() == m2.get_cache_key()
        True
        >>> m2.opt.etag = b'000'
        >>> m1.get_cache_key() == m2.get_cache_key()
        False
        >>> ignore = [OptionNumber.ETAG]
        >>> m1.get_cache_key(ignore) == m2.get_cache_key(ignore)
        True
        """

        options = []

        for option in self.opt.option_list():
            if option.number in ignore_options or (option.number.is_safetoforward() and option.number.is_nocachekey()):
                continue
            options.append((option.number, option.value))

        return (self.code, tuple(options))

    #
    # splitting and merging messages into and from message blocks
    #

    def _extract_block(self, number, size_exp):
        """Extract block from current message."""
        size = 2 ** (size_exp + 4)
        start = number * size
        if start < len(self.payload):
            end = start + size if start + size < len(self.payload) else len(self.payload)
            block = copy.deepcopy(self)
            block.payload = block.payload[start:end]
            block.mid = None
            more = True if end < len(self.payload) else False
            if block.code.is_request():
                block.opt.block1 = (number, more, size_exp)
            else:
                block.opt.block2 = (number, more, size_exp)
            return block

    def _append_request_block(self, next_block):
        """Modify message by appending another block"""
        if not self.code.is_request():
            raise ValueError("_append_request_block only works on requests.")

        block1 = next_block.opt.block1
        if block1.start == len(self.payload):
            self.payload += next_block.payload
            self.opt.block1 = block1
            self.token = next_block.token
            self.mid = next_block.mid
        else:
            raise error.NotImplemented()

    def _append_response_block(self, next_block):
        """Append next block to current response message.
           Used when assembling incoming blockwise responses."""
        if not self.code.is_response():
            raise ValueError("_append_response_block only works on responses.")

        block2 = next_block.opt.block2
        if block2.start != len(self.payload):
            raise error.NotImplemented()

        if next_block.opt.etag != self.opt.etag:
            raise error.ResourceChanged()

        self.payload += next_block.payload
        self.opt.block2 = block2
        self.token = next_block.token
        self.mid = next_block.mid

    def _generate_next_block2_request(self, response):
        """Generate a request for next response block.

        This method is used by client after receiving blockwise response from
        server with "more" flag set."""
        request = copy.deepcopy(self)
        request.payload = b""
        request.mid = None
        if response.opt.block2.block_number == 0 and response.opt.block2.size_exponent > DEFAULT_BLOCK_SIZE_EXP:
            new_size_exponent = DEFAULT_BLOCK_SIZE_EXP
            new_block_number = 2 ** (response.opt.block2.size_exponent - new_size_exponent)
            request.opt.block2 = (new_block_number, False, new_size_exponent)
        else:
            request.opt.block2 = (response.opt.block2.block_number + 1, False, response.opt.block2.size_exponent)
        del request.opt.block1
        del request.opt.observe
        return request

    def _generate_next_block1_response(self):
        """Generate a response to acknowledge incoming request block.

        This method is used by server after receiving blockwise request from
        client with "more" flag set."""
        response = Message(code=CHANGED, token=self.token)
        response.remote = self.remote
        if self.opt.block1.block_number == 0 and self.opt.block1.size_exponent > DEFAULT_BLOCK_SIZE_EXP:
            new_size_exponent = DEFAULT_BLOCK_SIZE_EXP
            response.opt.block1 = (0, True, new_size_exponent)
        else:
            response.opt.block1 = (self.opt.block1.block_number, True, self.opt.block1.size_exponent)
        return response

    #
    # the message in the context of network and addresses
    #

    @staticmethod
    def _build_request_uri(scheme, hostinfo, path, query):
        """Assemble path components as found in CoAP options into a URL. Helper
        for :meth:`get_request_uri`."""

        netloc = hostinfo

        # FIXME this should follow coap section 6.5 more closely
        query = "&".join(query)
        path = '/'.join(("",) + path) or '/'

        fragment = None
        params = ""  # are they not there at all?

        return urllib.parse.urlunparse((scheme, netloc, path, params, query, fragment))

    def get_request_uri(self):
        """The absolute URI this message belongs to.

        For requests, this is composed from the options (falling back to the
        remote). For responses, this is stored by the Request object not only
        to preserve the request information (which could have been kept by the
        requesting application), but also because the Request can know about
        multicast responses (which would update the host component) and
        redirects (FIXME do they exist?)."""

        # maybe this function does not belong exactly *here*, but it belongs to
        # the results of .request(message), which is currently a message itself.

        hostinfo = None
        host = None

        if self.code.is_response():
            proxyuri = self.requested_proxy_uri
            scheme = self.requested_scheme or 'coap'
            query = self.requested_query
            path = self.requested_path
        else:
            proxyuri = self.opt.proxy_uri
            scheme = self.opt.proxy_scheme or 'coap'
            query = self.opt.uri_query or ()
            path = self.opt.uri_path

        if self.code.is_response() and self.requested_hostinfo is not None:
            hostinfo = self.requested_hostinfo
        elif self.code.is_request() and self.opt.uri_host is not None:
            host = self.opt.uri_host
        elif self.code.is_request() and self.unresolved_remote is not None:
            hostinfo = self.unresolved_remote
        else:
            hostinfo = self.remote.hostinfo

        if self.code.is_request() and self.opt.uri_port is not None:
            port = self.opt.uri_port
        else:
            port = None

        if proxyuri is not None:
            return proxyuri

        if hostinfo is None and host is None:
            raise ValueError("Can not construct URI without any information on the set host")

        if hostinfo is None:
            hostinfo = hostportjoin(host, port)

        return self._build_request_uri(scheme, hostinfo, path, query)

    def set_request_uri(self, uri, *, set_uri_host=True):
        """Parse a given URI into the uri_* fields of the options.

        The remote does not get set automatically; instead, the remote data is
        stored in the uri_host and uri_port options. That is because name resolution
        is coupled with network specifics the protocol will know better by the
        time the message is sent. Whatever sends the message, be it the
        protocol itself, a proxy wrapper or an alternative transport, will know
        how to handle the information correctly.

        When ``set_uri_host=False`` is passed, the host/port is stored in the
        ``unresolved_remote`` message property instead of the uri_host option;
        as a result, the unresolved host name is not sent on the wire, which
        breaks virtual hosts but makes message sizes smaller."""

        parsed = urllib.parse.urlparse(uri, allow_fragments=False)

        if parsed.scheme != 'coap':
            self.opt.proxy_uri = uri
            return

        if parsed.username or parsed.password:
            raise ValueError("User name and password not supported.")

        # FIXME as with get_request_uri, this hould do encoding/decoding and section 6.5 etc

        if parsed.path not in ('', '/'):
            self.opt.uri_path = parsed.path.split('/')[1:]
        else:
            self.opt.uri_path = []
        if parsed.query:
            self.opt.uri_query = parsed.query.split('&')
        else:
            self.opt.uri_query = []

        if set_uri_host:
            if parsed.port:
                self.opt.uri_port = parsed.port
            self.opt.uri_host = parsed.hostname
        else:
            self.unresolved_remote = parsed.netloc



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""Module in which all meaningful numbers are collected. Most of the submodules
correspond to IANA registries."""

from . import constants, types, codes
from .constants import *
from .types import *
from .codes import *
from .optionnumbers import OptionNumber

__all__ = constants.__all__ + types.__all__ + codes.__all__ + ['OptionNumber']

media_types = {0: 'text/plain',
               40: 'application/link-format',
               41: 'application/xml',
               42: 'application/octet-stream',
               47: 'application/exi',
               50: 'application/json',
               51: 'application/json-patch+json', # draft-ietf-core-etch-04
               52: 'application/merge-patch+json', # draft-ietf-core-etch-04
               }
"""A map from CoAP-assigned integral codes to Internet media type descriptions."""

media_types_rev = {v:k for k, v in media_types.items()}



"""
Bit Writing Request/Response
------------------------------

TODO add ExtException to protocol
"""
from Bubot.Helpers.Coap.coap import Message, Code
from Bubot.Helpers.Helper import Helper
from Bubot.Helpers.ExtException import ExtException
import asyncio
from Bubot.Core.OcfMessage import OcfRequest, OcfResponse
import logging

# _logger = logging.getLogger('CoapProtocol')


class CoapProtocol:
    def __init__(self, server, **kwargs):
        # self.loop = loop
        self.transport = None
        self.server = server
        self.log = server.device.log
        self.multicast = kwargs.get('multicast', False)
        # self.on_con_made = loop.create_future()
        # self.on_con_lost = loop.create_future()

    def connection_made(self, transport):
        self.transport = transport
        # self.on_con_made.set_result(True)

    def datagram_received(self, data, remote):
        try:
            # self.log.debug(remote[1])
            message = Message.decode(data, remote)
        except Exception as e:
            logging.warning('datagram message encoding: {0} from {1}'.format(e, remote[0]))
            return
        code = message.code
        if code.is_request():
            handler_name = 'on_{0}'.format(code.name.lower())
            if not hasattr(self, handler_name):
                self.log.warning('handler not found {}'.format(handler_name))
                return
            try:
                message = OcfRequest.decode_from_coap(message, self.multicast)
            except Exception as e:
                logging.warning('OcfRequest message encoding: {0} from {1}'.format(e, remote[0]))
                return
            try:
                self.log.debug('{0} {1}'.format(
                    handler_name,
                    message.mid
                ))
                self.server.loop.create_task(getattr(self, handler_name)(message))
            except Exception as err:
                # self.server.OcfDriver.log.error('{0} {1}'.format(handler, e))
                answer = Message.generate_error(err, message)
                self.server.send_message(answer, remote)
                self.log.error(err)

        else:
            try:
                answer = self.server.answer[message.token]
            except KeyError:
                self.log.warning('request not found token={0} href={1} remote={2}'.format(
                    message.token, message.opt.uri_path, message.remote))  # надо сделать отбивку неожидаемых ответов
                return
            try:
                message = OcfResponse.decode_from_coap(message, self.multicast)
            except Exception as e:
                logging.warning('OcfResponse message encoding: {0} from {1}'.format(e, remote[0]))
                return

            if asyncio.isfuture(answer['response']):
                if code.is_successful():
                    self.log.debug('response {0}'.format(
                        message.token
                    ))
                    answer['response'].set_result(message)
                else:
                    self.log.debug('exception {0}'.format(
                        message.token
                    ))
                    try:
                        answer['response'].set_exception(Helper.loads_exception(message.cn))
                    except Exception as err:
                        answer['response'].set_exception(ExtException(
                            message='Load Exception error',
                            parent=err,
                            dump={'data': message.cn}
                        ))
                return
            elif callable(answer['response']):
                try:
                    self.log.debug('long response {0}'.format(
                        message.token
                    ))
                    self.server.loop.create_task(answer['response'](message, answer))
                except Exception as err:
                    self.log.error(err)
            # if isinstance(answer['response'], list):
            #     answer['response'].append(message)

        # self.server.bridge.on_datagram_received(data, addr)

    def error_received(self, exc):
        print('Error received:', exc)

    def connection_lost(self, exc):
        print("Connection closed")
        # self.on_con_lost.set_result(True)

    # async def on_content(self, message):
    #     try:
    #         answer = self.server.answer[message.mid]
    #     except KeyError:
    #         self.log.warning('request not found')
    #         return
    #     if asyncio.isfuture(answer['response']):
    #         answer['response'].set_result(message)
    #         return
    #     answer.append(message)

    async def on_get(self, message):
        try:
            answer_data = await self.server.device.on_get_request(message)
            answer = OcfResponse.generate_answer(answer_data, message)
            await self.server.send_answer(answer)
        except asyncio.CancelledError:
            pass
        except Exception as err:
            # self.server.OcfDriver.log.error('{0} {1}'.format(handler, e))
            self.log.error('on_get {0} {1} {2}: {3}'.format(
                self.server.device.__class__.__name__,
                message.operation,
                message.to.href,
                err))
            answer = OcfResponse.generate_error(err, message)
            await self.server.send_answer(answer)

    async def on_post(self, message):
        try:
            answer_data = await self.server.device.on_post_request(message)
            answer = OcfResponse.generate_answer(answer_data, message, code=Code.CHANGED)
            await self.server.send_answer(answer)
        except Exception as err:
            # self.log.debug(f'send error answer {message.to.href} {err}')
            answer = OcfResponse.generate_error(err, message)
            await self.server.send_answer(answer)



# This file is part of the Python aiocoap library project.
#
# Copyright (c) 2012-2014 Maciej Wasilak <http://sixpinetrees.blogspot.com/>,
#               2013-2014 Christian Amsüss <c.amsuess@energyharvesting.at>
#
# aiocoap is free software, this file is published under the MIT license as
# described in the accompanying LICENSE file.

"""Helpers for creating server-style applications in aiocoap

Note that these are not particular to aiocoap, but are used at different places
in aiocoap and thus shared here."""

import sys
import logging
import asyncio


class AsyncCLIDaemon:
    """Helper for creating daemon-style CLI prorgrams.

    Note that this currently doesn't create a Daemon in the sense of doing a
    daemon-fork; that could be added on demand, though.

    Subclass this and implement the :meth:`start` method as an async
    function; it will be passed all the constructor's arguments.

    When all setup is complete and the program is operational, return from the
    start method.

    Implement the :meth:`shutdown` coroutine and to do cleanup; what actually
    runs your program will, if possible, call that and await its return.

    Typical application for this is running ``MyClass.sync_main()`` in the
    program's ``if __name__ == "__main__":`` section."""

    def __init__(self, *args, **kwargs):
        self.initializing = asyncio.Task(self.start(*args, **kwargs))

    @classmethod
    def sync_main(cls, *args, **kwargs):
        """Run the application in an AsyncIO main loop, shutting down cleanly
        on keyboard interrupt."""
        logging.basicConfig(level=logging.DEBUG)
        loop = asyncio.get_event_loop()
        main = cls(*args, **kwargs)
        try:
            loop.run_until_complete(main.initializing)
            # This is the time when we'd signal setup completion by the parent
            # exiting in case of a daemon setup, or to any other process
            # management.
            logging.info("Application ready.")
            loop.run_forever()
        except KeyboardInterrupt:
            logging.info("Keyboard interupt received, shutting down")
            sys.exit(3)
        finally:
            if main.initializing.done() and main.initializing.exception():
                pass  # will raise from run_until_complete
            else:
                loop.run_until_complete(main.initializing)
                loop.run_until_complete(main.shutdown())
                loop.stop()
